; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     x87 Hardware Emulator                           *
; *                                                     *
; *     Copyright (C) 2002-2002 Norbert Juffa (_exp)    *
; *     Copyright (C) 2000-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   EM87

             INCLUDE SE.ASM

             .386
             .387

        if DPMIVersion

             EXTRN   __AHIncr:ABS

        endif

DATA         SEGMENT WORD PUBLIC USE16

; Externals

             EXTRN   ln2_hi: QWORD
             EXTRN   ln2_lo: QWORD

DATA         ENDS

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE,DS:DATA

             EXTRN   HaltError: NEAR

             PUBLIC  E087_ENTRY
             PUBLIC  E087_SHORTCUT

;-------------------------------------------------------------------------------
; Shortcut code 0xEE
;
; Cosine shortcut routine - move into F87H.ASM
;-------------------------------------------------------------------------------
_cos:        fcos
             jmp     $check

;-------------------------------------------------------------------------------
; Shortcut code 0xEC
;
; Sine shortcut routine - move into F87H.ASM
;-------------------------------------------------------------------------------
_sin:        fsin

$check:      wait                      ; check for unmasked exceptions
             fnstsw  ax                ; get FPU condition codes
             sahf                      ; transfer condition codes to CPU flags
             mov     ax, 207           ; error code 207, illegal float operation
             jp      HaltError         ; out of range, error

$exit:       ret

;-------------------------------------------------------------------------------
; Shortcut code 0xF2
;
; Arctan shortcut routine - move into F87H.ASM
;-------------------------------------------------------------------------------
_arctan:     fld1
             fpatan
             ret

;-------------------------------------------------------------------------------
; Shortcut code 0xF4
;
; Ln shortcut routine - move into F87H.ASM
;-------------------------------------------------------------------------------
_ln:         fldln2
             fxch
             fyl2x
             ret

;-------------------------------------------------------------------------------
; Shortcut code 0xFA
;
; Exp (e**x) shortcut routine - move into F87H.ASM
;-------------------------------------------------------------------------------
_exp:
             push    ds
             push    ax
             mov     ax, seg data
             mov     ds, ax
                               ; x
             fldl2e            ; log2(e) x
             fmul    st, st(1) ; z = x * log2(e) x
             frndint           ; int(z) x
             fld     ln2_hi    ; ln2_hi int(z) x
             fmul    st, st(1) ; int(z)*ln2_hi int(z) x
             fsubp   st(2), st ; int(z) x-int(z)*ln2_hi
             fld     ln2_lo    ; ln2_lo int(z) x-int(z)*ln2_hi
             fmul    st, st(1) ; int(z)*ln2_lo int(z) x-int(z)*ln2_hi
             fsubp   st(2), st ; int(z) (x-int(z)*ln2_hi)-int(z)*ln2_lo
             fxch    st(1)     ; (x-int(z)*ln2_hi)-int(z)*ln2_lo int(z)
             fldl2e            ; log2(e) (x-int(z)*ln2_hi)-int(z)*ln2_lo int(z)
             fmulp   st(1), st ; frac(z) int(z)
             f2xm1             ; 2^frac(z)-1 int(z)
             fld1              ; 1 2^frac(z)-1 int(z)
             faddp   st(1), st ; 2^frac(z) int(z)
             fscale            ; 2^z int(z)
             fstp    st(1)     ; 2^z = e^x

             pop     ax
             pop     ds

_nop:        ret

;-------------------------------------------------------------------------------
; E087_ENTRY:
;
; This interrupt routine is used to translate the original, emulated through
; interrupts 34-3D, floating point code into real FPU instructions.
;
; Interrupt 34..3B emulate FPU opcodes D8..DF
; Interrupt 3D emulates a stand-alone FWAIT
; Interrupt 3C is used to emulate opcodes with segment overrides
;
; From Ralf Brown:
;
; The generated code is CD 3C xy mm ....
; where xy is a modified ESC instruction and mm is the modR/M byte.
;
; The xy byte appears to be encoded as
;   s s 0 1 1 x x x  or  s s 0 0 0 x x x
;
; where      "ss" specifies the segment override:
;   00 -> DS: [3Eh (001 11 110)]
;   01 -> SS: [36h (001 10 110)]
;   10 -> CS: [2Eh (001 01 110)]
;   11 -> ES: [26h (001 00 110)]
;-------------------------------------------------------------------------------
E087_ENTRY   PROC    NEAR

             push    bp
             mov     bp, sp
             push    ax
             push    si
             push    ds
             test    byte ptr [bp + 7], 02h  ; pushed Interrupt flag
             jz      @@1

             sti

        if DPMIVersion

@@1:         mov     si, [bp + 2]
             mov     ax, [bp + 4]
             add     ax, __AHIncr
             mov     ds, ax

        else

@@1:         lds     si, [bp + 2]      ; load interrupt return address

        endif

             dec     si                ; back to get interrupt number
             mov     ax, [si]          ; get interrupt no[AL]/first byte[AH]
             dec     si                ; back to interrupt opcode
             sub     al, 34h           ; emulator uses 34..3D
             cmp     al, 09h
             ja      @@4               ; not an emulator call

             je      @@5               ; 3D: standalone wait

             cmp     al, 8             ; 3C: segment override
             je      @@6

             add     al, 0D8h          ; convert back to D8..DF

@@2:         mov     ah, al            ; AH will end up at higher address
             mov     al, 90h           ; NOP

@@3:         mov     [si], ax
             mov     bp, sp
             mov     [bp + 8], si      ; return to NOP

@@4:         pop     ds
             pop     si
             pop     ax
             pop     bp

             iret

@@5:         mov     ax, 9B90h         ; NOP + FWAIT
             jmp     @@3

@@6:         test    ah, 20h           ; 0010 0000 not valid, NOP
             jnz     @@4

             mov     al, ah            ; duplicate first byte
             and     ax, 07C0h         ; 0000 0fgh ab00 0000
             shr     al, 3             ;           000a b000
             xor     al, 18h
             add     ax, 0D826h        ; base + "seges"
             mov     [si + 2], ah
             jmp     @@2

E087_ENTRY   ENDP

;-------------------------------------------------------------------------------
; From Ralf Brown's Interrupt list:
;
; INT 3E - FLOATING POINT EMULATION - Borland LANGUAGES "SHORTCUT" CALL
; Notes: the two bytes following the INT 3E instruction are the subcode
;        and a NOP (90h), except for subcodes DCh and DEh, where the
;        second byte is a register count (01h-08h)
;
; Values for Borland floating-point shortcut subcode:
;
; Subcode  Function
;
;  0DCh    load 8086 stack with 8087 registers; overwrites the 10*N bytes at the
;            top of the stack prior to the INT 3E with the 8087 register contents
;
;  0DEh    load 8087 registers from top of 8086 stack; ST0 is furthest from top
;            of 8086 stack
;
;  0E0h    round TOS and R1 to single precision, compare, pop twice
;            returns AX=8087 status word, FLAGS=8087 condition bits
;
;  0E2h    round TOS and R1 to double precision, compare, pop twice
;            returns AX=8087 status word, FLAGS=8087 condition bits
;
;          Note: buggy in TPas 5.5 & TPas 6.0, because it sets the 8087
;            precision control field to the undocumented value 01h; this
;            results in actually rounding to single precision
;
;  0E4h    compare TOS/R1 with two POP's
;            returns FLAGS=8087 condition bits
;
;  0E6h    compare TOS/R1 with POP
;            returns FLAGS=8087 condition bits
;
;  0E8h    FTST (check TOS value)
;            returns FLAGS=8087 condition bits
;
;  0EAh    FXAM (check TOS value)
;            returns AX=8087 status word
;
;  0ECh    sine(ST0)
;  0EEh    cosine(ST0)
;  0F0h    tangent(ST0)
;  0F2h    arctangent(ST0)
;  0F4h    ST0 = ln(ST0)
;  0F6h    ST0 = log2(ST0)
;  0F8h    ST0 = log10(ST0)
;  0FAh    ST0 = e**ST0
;  0FCh    ST0 = 2**ST0
;  0FEh    ST0 = 10**ST0
;-------------------------------------------------------------------------------
             INCLUDE ALIGN.ASM

             dw      CODE:_sin
             dw      CODE:_cos
             dw      CODE:_nop
             dw      CODE:_arctan
             dw      CODE:_ln
             dw      CODE:_nop
             dw      CODE:_nop
             dw      CODE:_exp
             dw      CODE:_nop
             dw      CODE:_nop

E087_SHORTCUT:

             sti
             push    bx
             push    es
             push    bp
             mov     bp, sp
             les     bx, [bp + 6]
             movsx   bx, byte ptr es:[bx]
             add     word ptr [bp + 6], 2
             cmp     bl, 0ECh                ; zapped 8 entries?
             jb      @@1                     ; not a valid shortcut

             call    word ptr cs:[bx + E087_SHORTCUT]

@@1:         pop     bp
             pop     es
             pop     bx
             iret

             INCLUDE ALIGN.ASM

CODE         ENDS

             END