; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Multiple Precision Argument Reduction      *
; *                                                     *
; *     Copyright (C) 1991-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FRED

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   RealAdd     : NEAR
             EXTRN   RealMulNoChk: NEAR
             EXTRN   RealMul     : NEAR

; Publics

             PUBLIC  RealReduceMP

;-----------------------------------------------------------------------------
; Routine RealReduceMP realizes an extended precision argument reduction as
; needed in the calculation of trigonometric functions Sin, Cos, and Tan and
; in the Exp function. In the argument reduction for these functions there is
; the danger of a great loss of precision due to cancelling when subtracting
; numbers of nearly equal magnitude. The reduction usually calls for the
; subtraction of some multiple n of a constant c from the argument x. The
; precision can be increased if the constant is split into two parts, c1 and
; c2, with c1 being the most significant part and c2 being the least
; significant part and c1+c2=c with more bits of precision than there are
; bits in the mantissa of the floating point format used for the calculations.
; Instead of computing x - n*c, the calculation is done as x - n*c1 - n*c2,
; where it is important that n*c1 can be exactly represented in the floating
;point format used.
;
; INPUT:     DX:BX:AX  n
;            DI:SI:CX  x
;            CS:BP     Pointer to table containing two REAL constants c1,c2
;
; OUTPUT:    DX:BX:AX  x + n*c1 + n*c2
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;------------------------------------------------------------------------------

RealReduceMP PROC    NEAR

             push    dx                ; save
             push    bx                ;  n on
             push    ax                ;   stack
             push    di                ; save x
             push    si                ;  on
             push    cx                ;   stack
             mov     cx, cs:[bp + 0]   ; get c1
             mov     si, cs:[bp + 2]   ;  from constant table
             mov     di, cs:[bp + 4]   ;   pointed to by CS:BP
             call    RealMulNoChk      ; compute n*c1

             pop     cx                ; get
             pop     si                ;  back
             pop     di                ;   x
             call    RealAdd           ; compute x + n*c1

             pop     cx                ; get
             pop     si                ;  back
             pop     di                ;   n
             push    dx                ; save
             push    bx                ;  x + n*c1
             push    ax                ;   on stack
             mov     ax, cs:[bp + 6]   ; get
             mov     bx, cs:[bp + 8]   ;  c2 from
             mov     dx, cs:[bp + 10]  ;   constant table
             call    RealMul           ; compute n*c2

             pop     cx                ; get
             pop     si                ;  back
             pop     di                ;   x + n*c1
             jmp     RealAdd           ; compute x + n*c1 + n*c2 and return

RealReduceMP ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END