; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Fast Multiplication routine                     *
; *                                                     *
; *     Copyright (C) 1991-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   FASTMUL2

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

             PUBLIC  RealMulF
             PUBLIC  RealMulFNoChk
             PUBLIC  RealMulFNChk2
             PUBLIC  ShortMul
             PUBLIC  ShortMulRev

; DI:..:CL
; DX:BX:AX

ShortMulRev  PROC    NEAR

             xchg    ax, cx
             mov     bx, si
             xchg    dx, di

ShortMulRev  ENDP

ShortMul     PROC    NEAR

             push    bp              ; save TURBO-framepointer
             xchg    bx, di          ; BX = b1, DI = a2
             mov     bp, dx          ; get sign of multiplicant
             xor     bp, bx          ; compute sign of result
             and     bp, 8000h       ; mask out sign bit
             xchg    al, ch          ; save b3
             add     cl, ch          ; sum of biased exponents
             sbb     ch, ch          ; clear msb
             neg     ch              ;  and put possible overflow in CH
             or      cx, bp          ; zap in sign bit
             push    cx              ; save new exponent and sign bit
             xor     cx, cx          ; clear lo-bytes of a3 and b3
             or      dh, 80h         ; set implicit bit of multipicand
             or      bh, 80h         ; set implicit bit of multiplicator
             mov     si, dx          ; save a1
             mul     bx              ; b1 * a3
             mov     bp, ax          ; generate sticky byte = 0
             xchg    ax, dx          ; AX = msw of product
             xchg    ax, di          ; save msw of product, get a2
             mul     bx              ; b1 * a2
             xchg    ax, bx          ; save lsw of product, get b1
             xchg    dx, si          ; save msw of product, get a1
             add     bx, di          ; add product
             adc     si, cx          ;  to FPA
             mul     dx              ; b1 * a1
             add     ax, si          ; add product
             adc     dx, cx          ;  result in DX:AX:BX
             jmp     $end_mantiss    ; handle exponent

$zero_res:   jmp     $zero_prod2     ; result is 0

ShortMul     ENDP

             INCLUDE ALIGN.ASM

RealMulF     PROC    NEAR

             test    cl, cl          ; multiplicator = 0 ?
             jz      $zero_res       ; result will be 0

RealMulFNoChk PROC   NEAR

             test    al, al          ; multiplicand = 0 ?
             jz      $zero_res       ; result is zero

RealMulFNChk2 PROC   NEAR

             push    bp              ; save TURBO-framepointer
             xchg    bx, di          ; BX = b1, DI = a2
             mov     bp, dx          ; get sign of multiplicant
             xor     bp, bx          ; compute sign of result
             and     bp, 8000h       ; mask out sign bit
             xchg    al, ch          ; save b3
             add     cl, ch          ; sum of biased exponents
             sbb     ch, ch          ; clear msb
             neg     ch              ;  and put possible overflow in CH
             or      cx, bp          ; zap in sign bit
             push    cx              ; save new exponent and sign bit
             xor     cx, cx          ; clear lo-bytes of a3 and b3
             or      dh, 80h         ; set implicit bit of multipicand
             or      bh, 80h         ; set implicit bit of multiplicator

$full_mult:  xchg    al, ch          ; CH = b3, AL = 0
             push    bx              ; save b1
             push    dx              ; save a1
             mov     bp, dx          ; save a1
; b1 = BX
; b2 = SI
; b3 = CX
; a1 = DX
; a2 = DI
; a3 = AX
             mul     bx              ; b1 * a3
             xor     bx, bx          ; clear FPA
             xchg    ax, cx          ; get b3, save LSW (b1*a3)
             xchg    dx, bp          ; get a1, save MSW (b1*a3)
             mul     dx              ; a1 * b3
             add     cx, ax          ; add
             adc     bp, dx          ;  result
             adc     bx, bx          ;   to FPA
             mov     ax, si          ; b2
             mul     di              ; a2 * b2
             adc     cx, ax
             adc     bp, dx
             adc     bx, 0
             xor     cx, cx          ; FPA = CX:BX:BP
             xchg    ax, si          ; get b2
             pop     si              ; get a1
             mul     si              ; a1 * b2
             add     bp, ax          ; add
             adc     bx, dx          ;  result
             adc     cx, cx          ;   to FPA
             xchg    ax, di          ; get a2
             pop     di              ; get b1
             mul     di              ; a2 * b1
             add     bp, ax          ; add result
             xchg    ax, di          ; get a1
             xchg    cx, si          ; CX = b1
             mov     di, bx          ; FPA = SI:DI:BX
             mov     bx, bp          ;

$sqr_end:    adc     di, dx          ;  to   SI:DI:BX
             adc     si, 0           ;   FPA
             mul     cx              ; a1 * b1
             add     ax, di
             adc     dx, si          ; result in DX:AX:BX

$end_mantiss:pop     cx              ; CH = exponent  CL = sign
             xchg    ax, bx          ; DX:BX:AX = result
             sub     cx, 81h         ; compute new exponent-1

$div_end:    test    dx, dx          ; is mantissa normalized ?
             js      $add_sub_end    ; yes

             add     ax, ax          ; no, shift
             adc     bx, bx          ;  FPA 1 bit
             adc     dx, dx          ;   to the left
             dec     cx              ; adjust exponent

$add_sub_end:xor     si, si          ; load zero
             adc     ax, 80h         ; round
             adc     bx, si          ;  up
             adc     dx, si          ;   mantissa
             adc     cx, si          ; increment exponent if mantissa overfl.

$round_done: pop     bp              ; restore caller's frame pointer
             test    ch, 40h         ; test if (exponent-1) negative
             jnz     $zero_prod2     ; yes, underflow, return zero

             and     dh, 7fh         ; force MSB of mantissa to 0
             inc     cx              ; new exponent
             mov     al, cl          ; store exponent
             or      dh, ch          ; fill in sign bit
             shr     ch, 1           ; test if exponent overflow (> FFh)
             ret                     ; done

$zero_prod2: xor     ax, ax          ; load
             xor     bx, bx          ;  a
             xor     dx, dx          ;   zero
             ret                     ; done

RealMulFNChk2 ENDP
RealMulFNoChk ENDP
RealMulF      ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END