; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Comparison                                 *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   FP48CMP

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  CmpMantissa
             PUBLIC  CmpAbsValue
             PUBLIC  RCmp

;-------------------------------------------------------------------------------
; RealCmp compares two numbers in the TURBO-Pascal six byte real format. The
; flags are set as if two unsigned integers had been compared with the 80x86
; CMP instruction. If op1 > op2, neither carry nor zero flag are set. If
; op1 = op2 then the zero flag is set and if op1 < op2 then the carry flag is
; set. CmpAbsValue is an additional entry to the comparison routine that
; requires that both arguments are positive. Entry CmpMantissa is provided for
; arguments that are both positive and have identical exponents.
;
; INPUT:     DX:BX:AX  first number (op1)
;            DI:SI:CX  second number (op2)
;
; OUTPUT:    CF        set if op1 < op2
;            ZF        set if op1 = op2
;
; DESTROYS:  Flags
;-------------------------------------------------------------------------------

RCmp         PROC    FAR

             xor     dx, di            ; sign equal ?
             jns     $sign_eq          ; yes

             xor     dx, di            ; restore DX
             cmp     di, dx            ; negative number bigger, reverse compare

$exit1:      ret                       ; exit

$sign_eq:    xor     dx, di            ; restore DX
             jns     $cmp_pos          ; compare exponents and mantissas

             call    CmpAbsValue       ; call expo. & mant. compare for neg. num
             jz      $exit1            ; if numbers equal exit

             cmc                       ; else complement carry flag for neg. num
             ret                       ; done

$cmp_pos:    call    CmpAbsValue       ; compare two positive numbers
             ret                       ; done

RCmp         ENDP

             INCLUDE ALIGN.ASM

CmpAbsValue  PROC    NEAR

             cmp     al, cl            ; compare exponents
             jnz     $end              ; not equal, ->

             test    al, al            ; equal, but both zero ?
             jz      $end              ; yes

CmpAbsValue  ENDP

CmpMantissa  PROC    NEAR

             cmp     dx, di            ; compare
             jnz     $end              ;  mantissa bytes

             cmp     bx, si            ;   from msb
             jnz     $end              ;    to lsb and exit

             cmp     ah, ch            ;     if unequal or all equal

$end:        ret

CmpMantissa  ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END