; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Kernel Routines (Add,Sub,Mul,Div,Sqr)      *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   FP48KER

             .386

;-------------------------------------------------------------------------------
;
;  Turbo Pascal REAL floating-point format
;
;  47 46                            8 7       0
;  +--+------------------------------+--------+
;  |S |           Mantissa           |Exponent|
;  +--+------------------------------+--------+
;
;  47             31          15      7       0
;  +-------------+------------+------+--------+
;  |     DX      |     BX     |  AH  |   AL   |
;  +-------------+------------+------+--------+
;
;  47             31          15      7       0
;  +-------------+------------+------+--------+
;  |     DI      |     SI     |  CH  |   CL   |
;  +-------------+------------+------+--------+
;
;  value = 1^(-S) * Mantissa/2^40 * 2^(Exponent - 129)
;-------------------------------------------------------------------------------

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   HaltError: NEAR

; Publics

             PUBLIC  RealAdd
             PUBLIC  RealSub
             PUBLIC  RealMul
             PUBLIC  RealDiv
             PUBLIC  RealSqr
             PUBLIC  RealSqrNoChk
             PUBLIC  RealDivRev
             PUBLIC  RealMulNoChk
             PUBLIC  RealMulNChk2
             PUBLIC  RAdd
             PUBLIC  RSub
             PUBLIC  RMul
             PUBLIC  RDiv
             PUBLIC  RSqr
             PUBLIC  ROverflow

;-------------------------------------------------------------------------------
; RealAdd and RealSub are the routines for adding and subtracting two numbers
; in the Turbo Pascal 6 byte floating point format. They are practically ident-
; ical, since subtraction is implemented as an addition with a negated second
; addend. If underflow occurs, zero is returned. On overflow the carry flag
; will be set. The rounding of these routines complies with the IEEE "round to
; nearest or even" mode. Guard and sticky flags are therefore fully implemented.
;
; INPUT:     DX:BX:AX  first addend
;            DI:SI:CX  second addend
;
; OUTPUT:    DX:BX:AX  sum
;            CF        set if overflow occured, else cleared
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

AddExt       PROC    NEAR

$ret_second: xchg    ax, cx          ; load second addend
             mov     bx, si          ;  into DX:BX:AX (DX currently loaded)
             ret                     ; done

AddExt       ENDP

RealSub      PROC    NEAR

             xor     di, 8000h       ; negate second argument

RealSub      ENDP

RealAdd      PROC    NEAR

             cmp     cl, al          ; second addend bigger ?
             jae     $bigger         ; yes

             xchg    ax, cx          ; no,
             xchg    bx, si          ;  exchange
             xchg    dx, di          ;   addends

$bigger:     xchg    dx, di          ; DX = msb of second addend
             neg     al              ; smaller addend zero ?
             jz      $ret_second     ; yes, return other addend

             add     al, cl          ; compute difference of exponents
             cmp     al, 41          ; difference too big ?
             ja      $ret_second     ; yes, add/sub will not change bigger arg

             push    bp              ; save TURBO-Pascal frame pointer
             mov     bp, 0ff00h      ; load mask for msb
             and     bp, cx          ; save msb of second addend
             mov     ch, 80h         ; mask for sign bit
             and     ch, dh          ; sign bit of second addend
             push    cx              ; save sign and exponent
             xor     cx, di          ; test if operands have different sign
             pushf                   ; save sign indicator
             or      dh, 80h         ; set implicit bit in second addend
             xchg    dx, di          ; DX = msb of first addend
             or      dh, 80h         ; set implicit bit in first addend
             xor     cx, cx          ; set guard and sticky bytes to 0

             xchg    al, ch          ; DX:BX:AX = mantissa, CH = shift counter

$test_shift: cmp     ch, 4           ; less than 4 bit shifts necessary ?
             jb      $bit_shift      ; yes, do it one bit at a time

             cmp     ch, 8           ; between 4 and 7 bit shifts necessary ?
             jb      $4bit_shift     ; yes, do 4 bit shift first

             cmp     ch, 16          ; between 8 and 15 bit shifts necessary ?
             jb      $byte_shift     ; yes, do byte shift first

             or      cl, al          ; accumulate
             or      cl, ah          ;  sticky byte
             xchg    ax, dx          ; shift
             xchg    ax, bx          ;  mantissa 16 bits
             xor     dx, dx          ;   to the right
             sub     ch, 16          ; decrement shift counter by 16
             jmp     $test_shift     ; test remaining shifts

$byte_shift: or      cl, al          ; accumulate sticky byte
             mov     al, ah          ; shift
             mov     ah, bl          ;  mantissa
             mov     bl, bh          ;   eight
             mov     bh, dl          ;    bits to
             movzx   dx, dh          ;     the right
             test    ch, 4           ; 4 bit shift possible ?
             jz      $bit_shift      ; no, try single bit shifts

$4bit_shift: neg     cl              ; set sticky flag = FFh
             sbb     cl, cl          ;  if <> 0 before
             or      cl, al          ; accumulate
             and     cl, 0fh         ;  sticky flag
             shr     dx, 1           ; shift
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   1 bit to the right
             shr     dx, 1           ; shift
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   1 bit to the right
             shr     dx, 1           ; shift
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   1 bit to the right
             shr     dx, 1           ; shift
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   1 bit to the right

$bit_shift:  and     ch, 3           ; compute number of single bit shifts
             jz      $shift_done     ; no shifts necessary, mantissas aligned

             neg     cl              ; set sticky flag to FFh
             sbb     cl, cl          ;  if <> 0 before

             INCLUDE ALIGN.ASM

$bit_loop:   shr     dx, 1           ; shift
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   1 bit to the right
             adc     cl, cl          ; accumulate sticky byte
             dec     ch              ; decrement shift counter
             jnz     $bit_loop       ; until shift counter zero

$shift_done: popf                    ; signs of addends different ?
             js      $subtract       ; sign of addends differ

             add     ax, bp          ; add
             adc     bx, si          ;  mantissas
             adc     dx, di          ;   of two addends
             mov     bp, cx          ; get sticky byte
             pop     cx              ; get exponent and sign
             jnc     $no_overf       ; no mantissa overflow

             shr     dx, 1           ; divide
             rcr     bx, 1           ;  mantissa
             rcr     ax, 1           ;   by two
             inc     cx              ; adjust exponent

$no_overf:   dec     cx              ; exponent-1
             jmp     $add_sub_end    ; do rounding

$ret_first:  pop     bp              ; restore TURBO-Pascal frame pointer
             ret                     ; done

$subtract:   xchg    ax, bp          ; exchange
             xchg    bx, si          ;  addends
             xchg    dx, di          ;   for correct order
             neg     cx              ; set carry if sticky byte <> 0
             sbb     ax, bp          ; subtract
             sbb     bx, si          ;  the two
             sbb     dx, di          ;   mantissas
             mov     bp, cx          ; get sticky byte
             pop     cx              ; get exponent and sign of result
             jnc     $no_negate      ; no negative result

             xor     ch, 80h         ; result has other sign than 2. addend
             not     dx              ; negate
             not     bx              ;  number
             neg     ax              ;   in
             sbb     bx, -1          ;    DX:BX:AX
             sbb     dx, -1          ;     "

$no_negate:  js      $no_overf       ; mantissa normalized

             jz      $test_z1        ; first mantissa word is zero

             INCLUDE ALIGN.ASM

$shift_l:    dec     cx              ; adjust exponent
             add     ax, ax          ; multiply
             adc     bx, bx          ;  mantissa
             adc     dx, dx          ;   by two
             jns     $shift_l        ; normalized? no

             dec     cx              ; exponent-1
             jmp     $add_sub_end    ; do rounding

$test_z1:    xchg    bx, ax          ; do a 16-bit
             xchg    dx, ax          ;  left shift of the mantissa
             sub     cx, 16          ; adjust exponent
             test    dx, dx          ; first mantissa word zero?
             jg      $shift_l        ; not zero, no sign

             js      $no_overf       ; mantissa normalized

             xchg    dx, bx          ; shift mantissa 16 bits left (AX=0)
             sub     cx, 16          ; adjust exponent
             test    dx, dx          ; first mantissa word zero ?
             jg      $shift_l        ; not zero, no sign

             js      $no_overf       ; mantissa normalized

             pop     bp              ; mantissa zero, return DX:BX:AX=0
             ret                     ; done

RealAdd      ENDP

;-------------------------------------------------------------------------------
; RealMul multiplies two numbers in the Turbo Pascal 6 byte floating point
; format. If underflow occurs, zero is returned. On overflow the carry flag
; will be set. The routine multiplies the mantissas by computing nine partial
; products using the 80x86 MUL instruction. RealMulNoChk is the same routine
; as RealMul but does not check the operand in DI:SI:CX for zero. The fastest
; multiplication routine, RealMulNChk2, does not check either operand for zero.
; The rounding of this routine complies with IEEE "round to nearest or even"
; mode. For this purpose, guard and sticky flags are implemented.
;
; INPUT:     DX:BX:AX  multiplicand
;            DI:SI:CX  multiplicator
;
; OUTPUT:    DX:BX:AX  product
;            CF        set if overflow occured, else cleared
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

             INCLUDE ALIGN.ASM

RealMul      PROC    NEAR

             test    cl, cl          ; multiplicator = 0 ?
             jz      $zero_res       ; result will be 0

RealMulNoChk PROC    NEAR

             test    al, al          ; multiplicand = 0 ?
             jz      $zero_res       ; result is zero

RealMulNChk2 PROC    NEAR

             push    bp              ; save TURBO-framepointer
             xchg    bx, di          ; BX = b1, DI = a2
             mov     bp, dx          ; get sign of multiplicant
             xor     bp, bx          ; compute sign of result
             and     bp, 8000h       ; mask out sign bit
             xchg    al, ch          ; save b3
             add     cl, ch          ; sum of biased exponents
             sbb     ch, ch          ; clear msb
             neg     ch              ;  and put possible overflow in CH
             or      cx, bp          ; zap in sign bit
             push    cx              ; save new exponent and sign bit
             xor     cx, cx          ; clear lo-bytes of a3 and b3
             or      dh, 80h         ; set implicit bit of multipicand
             or      bh, 80h         ; set implicit bit of multiplicator
             test    si, si          ; b2 = 0 ?
             jz      $test_short     ; yes, test if b3 = 0

             test    di, di          ; a2 = 0 ?
             jnz     $full_mult      ; no, use full multiplication

             test    ah, ah          ; a3 = 0 ?
             jnz     $full_mult      ; no, use full multiplication

             xchg    ah, al          ; swap a3 and b3
             xchg    di, si          ; swap a2 and b2
             xchg    dx, bx          ; swap a1 and b1

$test_short: test    al, al          ; b3 = 0 ?
             jnz     $full_mult      ; no, use full multiplication

             mov     si, dx          ; save a1
             mul     bx              ; b1 * a3
             mov     bp, ax          ; generate sticky byte = 0
             xchg    ax, dx          ; AX = msw of product
             xchg    ax, di          ; save msw of product, get a2
             mul     bx              ; b1 * a2
             xchg    ax, bx          ; save lsw of product, get b1
             xchg    dx, si          ; save msw of product, get a1
             add     bx, di          ; add product
             adc     si, cx          ;  to FPA
             mul     dx              ; b1 * a1
             add     ax, si          ; add product
             adc     dx, cx          ;  result in DX:AX:BX
             jmp     $end_mantiss    ; handle exponent

$zero_res:   jmp     $zero_prod2     ; result is 0

             INCLUDE ALIGN.ASM

$full_mult:  push    bx              ; save b1
             push    di              ; save a2
             push    si              ; save b2
             push    dx              ; save a1
             push    bx              ; save b1
             mov     bx, cx          ; clear most significant word of FPA
             xchg    al, ch          ; CH = b3, AL = 0
             mov     bp, ax          ; a3
             mov     al, ch          ; b3
             mul     ah              ; a3 * b3
             xchg    ax, di          ; store to FPA, get a2
             mul     cx              ; a2 * b3
             add     di, ax          ; add result
             adc     dx, bx          ;  to FPA
             xchg    ax, dx          ;   and
             xchg    ax, si          ;    get b2
             mul     bp              ; a3 * b2
             add     di, ax          ; add result
             adc     si, dx          ;  to
             adc     bx, bx          ;   FPA
             xchg    ax, bp          ; get a3
             mov     bp, di          ; generate sticky flag
             xor     di, di          ; FPA = DI:BX:SI
             pop     dx              ; get b1
             mul     dx              ; a3 * b1
             add     si, ax          ; add result to
             adc     bx, dx          ;  FPA, no overflow possible
             xchg    ax, cx          ; b3
             pop     cx              ; a1
             mul     cx              ; a1 * b3
             add     si, ax          ; add
             adc     bx, dx          ;  result to
             adc     di, di          ;   FPA
             pop     ax              ; b2
             pop     dx              ; a2
             push    dx              ; save a2
             push    ax              ; save b2
             mul     dx              ; a2 * b2
             add     si, ax          ; add
             adc     bx, dx          ;  result
             adc     di, 0           ;   to FPA
             or      bp, si          ; accumulate sticky flag
             xor     si, si          ; FPA = SI:DI:BX
             pop     ax              ; b2
             mul     cx              ; a1 * b2
             add     bx, ax          ; add
             adc     di, dx          ;  result
             adc     si, si          ;   to FPA
             pop     ax              ; a2
             pop     dx              ; get b1
             push    dx              ; save b1
             mul     dx              ; a2 * b1
             add     bx, ax          ; add -------+
             pop     ax              ; get b1     !

$sqr_end:    adc     di, dx          ;  result  <-+
             adc     si, 0           ;   to FPA
             mul     cx              ; a1 * b1
             add     ax, di          ; add result to FPA
             adc     dx, si          ;  DX:AX:BX = result

$end_mantiss:pop     cx              ; CH = exponent  CL = sign
             xchg    ax, bx          ; DX:BX:AX = result
$div_end:    js      $$1

             add     ax, ax          ; no, shift
             adc     bx, bx          ;  FPA 1 bit
             adc     dx, dx          ;   to the left
             dec     cx              ; adjust exponent

$$1:         sub     cx, 81h

$add_sub_end:add     ax, 80h         ; round
             adc     bx, 0           ;  up
             adc     dx, 0           ;   mantissa
             adc     cx, 0           ; increment exponent if mantissa overfl.

             test    al, al          ; tie case ?
             jz      $tie_case       ; tie case possible if sticky = 0, too

$round_done: pop     bp              ; restore caller's frame pointer
             test    ch, 40h         ; test if (exponent-1) negative
             jnz     $zero_prod2     ; yes, underflow, return zero

             inc     cx              ; new exponent
             mov     al, cl          ; store exponent
             and     dh, 7fh         ; force MSB of mantissa to 0
             or      dh, ch          ; fill in sign bit
             shr     ch, 1           ; test if exponent overflow (> FFh)
             ret                     ; done

$zero_prod2: xor     ax, ax          ; load
             xor     bx, bx          ;  a
             cwd                     ;   zero
             ret                     ; done

$tie_case:   test    bp, bp          ; sticky flag = 0 (tie case) ?
             jnz     $round_done     ; no, round up was correct

             and     ah, 0feh        ; tie case, make mantissa even
             jmp     $round_done     ; IEEE rounding done

RealMulNChk2 ENDP
RealMulNoChk ENDP
RealMul      ENDP

             INCLUDE ALIGN.ASM

;-------------------------------------------------------------------------------
; RealSqr computes the square of a number in the Turbo Pascal 6-byte floating
; point format. If underflow occurs, zero is returned. On overflow the carry
; flag will be set. Since squaring allows for some optimizations in code when
; compared with normal multiplication, RealSqr is implemented as a self con-
; tained routine and not as a call to RealMul. The routine exits thru RealMul.
; RealSqrNoChk does not check the argument for zero before squaring. Rounding
; complies with the IEEE "round to nearest or even" mode, so guard and sticky
; flags are provided.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  square of argument
;            CF      set if overflow occured, else cleared
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RealSqr      PROC    NEAR

             test    al, al          ; argument = 0 ?
             jz      $zero_prod2     ; result is zero

RealSqrNoChk PROC    NEAR

             xor     cx, cx          ; clear register
             xchg    cl, al          ; exponent in CL, AL = 0
             add     cx, cx          ; new exponent, sign always positive (0)
             push    bp              ; save TURBO-Pascal frame pointer
             push    cx              ; save sign and exponent
             or      dh, 80h         ; set implicit bit of argument
             mov     si, ax          ; a2 and
             or      si, bx          ;  a3 = 0 ?
             jnz     $full_sqr       ; no, do full multiplication

             mov     ax, dx          ; load a1
             mul     dx              ; a1 * a1
             test    dx, dx
             jmp     $end_mantiss    ; result in DX:AX:BX

             INCLUDE ALIGN.ASM

$full_sqr:   push    bx              ; save a2
             xor     di, di          ; load zero
             mov     cx, dx          ; save a1
             mov     bp, ax          ; save a3
             mov     al, ah          ; load a3
             mul     al              ; a3 * a3
             xchg    ax, bx          ; save product, get a2
             mul     bp              ; a2 * a3
             xchg    ax, bp          ; get a3, BP = save lo-word a2*a3
             mov     si, dx          ; save hi-word a2*a3
             add     bx, bp          ; add a3*a3 to
             adc     si, di          ;  a2*a3 (result in SI:BX, no overflow)
             add     bp, bx          ; add a2*a3 lo-word to result
             mov     bx, di          ; BX = 0
             adc     si, dx          ; add a2*a3 hi-word
             adc     di, di          ;  to result (DI:SI:BP)
             xchg    di, bx          ; FPA = DI:BX:SI, BP = sticky byte
             mul     cx              ; a1 * a3
             add     si, ax          ; add product
             adc     bx, dx          ;  to FPA (no overflow possible)
             add     si, ax          ; add
             adc     bx, dx          ;  product to
             adc     di, di          ;   FPA another time
             pop     ax              ; get a2
             push    ax              ; save a2
             mul     ax              ; a2 * a2
             add     si, ax          ; add
             adc     bx, dx          ;  product to
             adc     di, 0           ;   FPA
             or      bp, si          ; accumulate sticky byte
             xor     si, si          ; FPA = SI:DI:BX
             pop     ax              ; get a2
             mul     cx              ; a1 * a2
             add     bx, ax          ; add
             adc     di, dx          ;  resulting
             adc     si, si          ;   product
             add     bx, ax          ;    to FPA twice
             mov     ax, cx          ; AX = CX = a1
             jmp     $sqr_end        ; exit thru REAL_MUL

RealSqrNoChk ENDP
RealSqr      ENDP

             INCLUDE ALIGN.ASM

;-------------------------------------------------------------------------------
; RealDiv divides two numbers in the Turbo Pascal 6 byte floating point
; format. If underflow occurs, zero is returned. On overflow the carry flag
; will be set. The routine exits through the REAL_MUL routine. It makes use
; of the 80x86 DIV instruction in an estimate and correct algorithm. In each
; of the three steps, an estimation of a part of the quotient is produced by
; dividing the first 32 bits of the current remainder by the first 16 bits of
; the divisor using a machine instruction. Then the divisor is multiplied by
; the result and this product subtracted from the current remainder. If the sum
; is negative, the partial quotient must be decremented until the new remainder
; is positive. RealDivRev is an additional routine which exchanges the operands
; before performing the division. The rounding provided complies with IEEE
; "round to nearest or even" mode. For this purpose, guard and sticky flags
; are implemented.
;
; INPUT:     DX:BX:AX  dividend
;            DI:SI:CX  divisor
;
; OUTPUT:    DX:BX:AX  quotient
;            CF        set if overflow occured, else cleared
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RealDivRev   PROC    NEAR

             xchg    ax, cx          ; exchange
             xchg    bx, si          ;  divisor and
             xchg    dx, di          ;   dividend

RealDivRev   ENDP

RealDiv      PROC    NEAR

             test    al, al          ; dividend = 0 ?
             jz      $zero_prod2     ; result is zero

             push    bp              ; save TURBO-Pascal framepointer
             mov     bp, dx          ; get msw of dividend
             xor     bp, di          ; xor with msw of divisor to make sign
             and     bp, 8000h       ; isolate sign bit of result
             or      dh, 80h         ; set implicit bit in dividend
             xchg    dx, di          ; DX = divisor msw, DI = dividend msw
             or      dh, 80h         ; set implicit bit in divisor
             sub     al, cl          ; subtract exponents ----------+
             mov     cl, 0           ; clear lsb of divisor lsw     |
             push    si              ;  save divisor middle word    |
             push    cx              ;   and lsw on stack           |
             mov     cx, bp          ; get sign                     |
             xchg    al, cl          ; AL = 0, CL = new exponent    |
             sbb     ch, al          ; put carry here <-------------+

             add     cx,101h
             mov     bp, sp          ; access divisor on stack via BP
             sub     bp, 6           ; leave room for three pushes
             shr     di, 1           ; divide dividend
             rcr     bx, 1           ;  by 2 to prevent
             rcr     ax, 1           ;   an overflow condition

             INCLUDE ALIGN.ASM

$divide_loop:push    cx              ; save sign & exponent resp. part. quot.
             mov     cx, dx          ; get msw of divisor
             xchg    ax, bx          ; create new dividend
             xchg    ax, di          ;  by shifting remainder
             xchg    ax, si          ;   16 bits to the left
             cmp     cx, si          ; overflow possible on division ?
             je      $div_overfl     ; yes

             mov     dx, si          ; get msw of dividend
             xchg    ax, di          ; second word of dividend
             div     cx              ; compute partial quotient
             xor     si, si          ; subtract product of divisor high word
             mov     di, dx          ;  and partial quotient from dividend

$comp_rem:   xchg    ax, cx          ; AX = divisor high word, CX = quotient
             push    ax              ; save divisor high word
             mov     ax, [bp+8]      ; get middle word of divisor
             mul     cx              ; multiply by partial quotient
             sub     bx, ax          ; subtract the product of
             sbb     di, dx          ;  divisor middle word and partial
             sbb     si, 0           ;   quotient from dividend
             mov     ax, [bp + 6]    ; get lsw of divisor
             mul     cx              ; multiply by partial quotient
             neg     ax              ; subtract the product
             sbb     bx, dx          ;   of divisor LSW
             sbb     di, 0           ;     and partial
             sbb     si, 0           ;      quotient from dividend
             pop     dx              ; get back msw of divisor
             jz      $sub_ok         ; remainder must be positive

$add_twice:  dec     cx              ; quotient to high, decrement it
             add     ax, [bp + 6]    ; adjust
             adc     bx, [bp + 8]    ;  quotient and
             adc     di, dx          ;   remainder
             jnc     $add_twice      ;     until remainder positive

$sub_ok:     cmp     bp, sp          ; two partial quotients saved already ?
             jne     $divide_loop    ; no, continue (carry set !!!)

             mov     bp, ax          ; accumulate
             or      bp, bx          ;  sticky
             or      bp, di          ;   byte
             xchg    ax, cx          ; get last partial quotient
             pop     bx              ; get other
             pop     dx              ;  partial quotients

             pop     cx              ; get sign and exponent
             add     sp, 4           ; remove saved divisor from stack
             test    dx, dx
             jmp     $div_end        ; normalize mantissa and round

$div_overfl: xor     si, si          ; remainder - 10000h * divisor
             add     di, cx          ; remainder -
             adc     si, si          ;  FFFFh * divisor
             mov     ax, -1          ; quotient = FFFFh
             jmp     $comp_rem       ; continue computation of remainder

RealDiv      ENDP

             INCLUDE ALIGN.ASM

RAdd         PROC    FAR

             call    RealAdd         ; perform addition
             jc      ROverflow       ; overflow error

             ret

RAdd         ENDP

             INCLUDE ALIGN.ASM

RSub         PROC    FAR

             call    RealSub         ; perform subtraction
             jc      ROverflow       ; overflow error

             ret

RSub         ENDP

             INCLUDE ALIGN.ASM

RSqr         PROC    FAR

             call    RealSqr         ; perform squaring
             jc      ROverflow       ; overflow error

             ret

RSqr         ENDP

             INCLUDE ALIGN.ASM

RMul         PROC    FAR

             call    RealMul         ; perform multiplication
             jc      ROverflow       ; overflow error

             ret

RMul         ENDP

             INCLUDE ALIGN.ASM

RDiv         PROC    FAR

             test    cl, cl          ; divisor zero ?
             jz      RDivZero        ; yes, error

             call    RealDiv         ; perform division
             jc      ROverflow       ; overflow error

             ret

RDiv         ENDP

ROverflow:   mov     ax, 205         ; error code 205 (fp overflow)
             jmp     HaltError       ; execute error handler

RDivZero:    mov     ax, 200         ; error code 200 (division by zero)
             jmp     HaltError       ; execute error handler

             INCLUDE ALIGN.ASM

CODE         ENDS

             END